#!/usr/bin/python
# -*- coding: utf-8 -*-

from decimal import Decimal
from datetime import datetime, date, timedelta
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

import pytest
from currency_converter import (CurrencyConverter, S3CurrencyConverter,
                                RateNotFoundError, ECB_URL,
                                SINGLE_DAY_ECB_URL)

c0 = CurrencyConverter()
c1 = CurrencyConverter(fallback_on_missing_rate=True)
c2 = CurrencyConverter(fallback_on_wrong_date=True)
c3 = CurrencyConverter(fallback_on_missing_rate=True, fallback_on_wrong_date=True,
                       fallback_on_missing_rate_method="linear_interpolation")
c4 = CurrencyConverter(fallback_on_missing_rate=True, fallback_on_wrong_date=True,
                       fallback_on_missing_rate_method="last_known")
c5 = CurrencyConverter('./currency_converter/eurofxref-hist.zip')

converters = [c0, c1, c2, c3, c5, c4]
converters_with_missing_rate_fallback = [c1, c3, c4]
converters_with_wrong_date_fallback = [c2, c3, c4]
converters_without_missing_rate_fallback = [c0, c2, c5]
converters_without_wrong_date_fallback = [c0, c1, c5]


@pytest.fixture
def fallback_with_linear_interpolation():
    return c3


@pytest.fixture
def fallback_with_last_known():
    return c4


@pytest.fixture
def decimal_converter():
    return CurrencyConverter(decimal=True)


def equals(a, b):
    return abs(b - a) < 1e-5


class TestRates(object):

    @pytest.mark.parametrize('c', converters)
    def test_convert(self, c):
        assert equals(c.convert(10, 'EUR', 'USD', date(2013, 3, 21)), 12.91)
        assert equals(c.convert(10, 'EUR', 'USD', date(2014, 3, 28)), 13.758999)
        assert equals(c.convert(10, 'USD', 'EUR', date(2014, 3, 28)), 7.26797)

    @pytest.mark.parametrize('c', converters)
    def test_convert_with_datetime(self, c):
        assert equals(c.convert(10, 'EUR', 'USD', datetime(2013, 3, 21)), 12.91)
        assert equals(c.convert(10, 'EUR', 'USD', datetime(2014, 3, 28)), 13.758999)
        assert equals(c.convert(10, 'USD', 'EUR', datetime(2014, 3, 28)), 7.26797)

    @pytest.mark.parametrize('c', converters)
    def test_convert_to_ref_currency(self, c):
        assert c.convert(10, 'EUR') == 10.
        assert c.convert(10, 'EUR', 'EUR') == 10.

    def test_decimal_converter(self, decimal_converter):
        dc = decimal_converter
        assert dc.convert(10, 'EUR', 'USD', date(2013, 3, 21)) == Decimal('12.910')
        assert dc.convert(10, 'EUR', 'USD', date(2014, 3, 28)) == Decimal('13.7590')
        assert dc.convert(10, 'USD', 'EUR', date(2014, 3, 28)) == Decimal('7.2679700559'
                                                                          '63369430917944618')
        assert dc.convert(10, 'EUR') == Decimal('10')
        assert dc.convert(10, 'EUR', 'EUR') == Decimal('10')


class TestErrorCases(object):

    @pytest.mark.parametrize('c', converters)
    def test_wrong_currency(self, c):
        with pytest.raises(ValueError):
            c.convert(1, 'AAA')

    @pytest.mark.parametrize('c', converters_without_missing_rate_fallback)
    def test_convert_with_missing_rate(self, c):
        with pytest.raises(RateNotFoundError):
            c.convert(10, 'BGN', date=date(2010, 11, 21))

    @pytest.mark.parametrize('c', converters_with_missing_rate_fallback)
    def test_convert_fallback_on_missing_rate(self, c):
        assert equals(c1.convert(10, 'BGN', date=date(2010, 11, 21)), 5.11299)

    @pytest.mark.parametrize('c', converters_without_wrong_date_fallback)
    def test_convert_with_wrong_date(self, c):
        with pytest.raises(RateNotFoundError):
            c.convert(10, 'EUR', 'USD', date=date(1986, 2, 2))

    @pytest.mark.parametrize('c', converters_with_wrong_date_fallback)
    def test_convert_fallback_on_wrong_date(self, c):
        assert equals(c.convert(10, 'EUR', 'USD', date=date(1986, 2, 2)), 11.789)

    def test_fallback_methds(self,
                             fallback_with_linear_interpolation,
                             fallback_with_last_known):
        li = fallback_with_linear_interpolation
        ln = fallback_with_last_known
        assert equals(li.convert(10, 'USD', date=date(2019, 12, 8)), 9.02418)
        assert equals(ln.convert(10, 'USD', date=date(2019, 12, 8)), 9.01388)


def last_n_days(n):
    return [date.today() - timedelta(days=d)
            for d in reversed(range(n + 1))]


class TestAttributes(object):

    @pytest.mark.parametrize('c', converters)
    def test_bounds(self, c):
        assert c.bounds['USD'][0] == date(1999, 1, 4)
        assert c.bounds['BGN'][0] == date(2000, 7, 19)
        assert c.bounds['EUR'][0] == date(1999, 1, 4)

        assert c.bounds['USD'][1] in last_n_days(7)
        assert c.bounds['BGN'][1] in last_n_days(7)
        assert c.bounds['EUR'][1] in last_n_days(7)

    @pytest.mark.parametrize('c', converters)
    def test_currencies(self, c):
        assert len(c.currencies) == 42
        assert 'EUR' in c.currencies


class TestCustomObject(object):

    c = CurrencyConverter(currency_file=None,
                          fallback_on_wrong_date=True,
                          fallback_on_missing_rate=True)

    c.load_lines(StringIO('''\
    Date,USD,AAA,
    2014-03-29,2,N/A
    2014-03-27,6,0
    2014-03-23,18,N/A
    2014-03-22,N/A,0'''))

    def test_convert(self):
        assert equals(self.c.convert(10, 'EUR', 'USD'), 20)
        assert equals(self.c.convert(10, 'USD', 'EUR'), 5)

    def test_fallback_date(self):
        # Fallback to 2014-03-29 rate of 2
        assert equals(self.c.convert(10, 'EUR', 'USD', date(2015, 1, 1)), 20)
        assert equals(self.c.convert(10, 'USD', 'EUR', date(2015, 1, 1)), 5)

        # Fallback to 2014-03-23 rate of 18
        assert equals(self.c.convert(10, 'EUR', 'USD', date(2012, 1, 1)), 180)
        assert equals(self.c.convert(10, 'USD', 'EUR', date(2012, 1, 1)), 0.555555)

    def test_fallback_rate(self):
        # Fallback rate is the average between 2 and 6, so 4
        assert equals(self.c.convert(10, 'EUR', 'USD', date(2014, 3, 28)), 40)
        assert equals(self.c.convert(10, 'USD', 'EUR', date(2014, 3, 28)), 2.5)

        # Fallback rate is the weighted mean between 6 (d:1) and 18 (d:3), so 9
        assert equals(self.c.convert(10, 'EUR', 'USD', date(2014, 3, 26)), 90)
        assert equals(self.c.convert(10, 'USD', 'EUR', date(2014, 3, 26)), 1.11111)

    def test_attributes(self):
        assert self.c.currencies == set(['EUR', 'USD', 'AAA'])
        assert self.c.bounds == {
            'USD': (date(2014, 3, 23), date(2014, 3, 29)),
            'AAA': (date(2014, 3, 22), date(2014, 3, 27)),
            # Max of previous ranges
            'EUR': (date(2014, 3, 22), date(2014, 3, 29))
        }


def test_single_day_file():
    c = CurrencyConverter(SINGLE_DAY_ECB_URL)
    assert len(c.currencies) == 33


def skip_test_on_error(msg):
    def decorator(test_function):
        def wrapped_test_function(*args, **kwargs):
            try:
                test_function(*args, **kwargs)
            except Exception as e:  # will not catch AssertionError
                if msg in str(e):
                    pytest.skip("Skipped because {0!r} found".format(msg))
                else:
                    raise
        return wrapped_test_function
    return decorator


class TestCustomSource(object):

    def test_local_zip_file(self):
        c = CurrencyConverter('./currency_converter/eurofxref-hist.zip')
        assert len(c.currencies) == 42
        assert equals(c.convert(10, 'EUR', 'USD', date(2013, 3, 21)), 12.91)

    def test_remote_zip_file(self):
        c = CurrencyConverter(ECB_URL)
        assert len(c.currencies) == 42
        assert equals(c.convert(10, 'EUR', 'USD', date(2013, 3, 21)), 12.91)

    def test_local_clear_file(self):
        c = CurrencyConverter('./currency_converter/test_single_day.csv')
        assert len(c.currencies) == 32

    @skip_test_on_error('CERTIFICATE_VERIFY_FAILED')
    def test_remote_clear_file(self):
        c = CurrencyConverter('https://raw.githubusercontent.com/alexprengere'
                              '/currencyconverter/master/currency_converter/test_single_day.csv')
        assert len(c.currencies) == 32


class TestS3(object):

    def test_S3_currency_file_required(self):
        with pytest.raises(TypeError):
            S3CurrencyConverter()

    def test_S3_currency_file_needs_get_contents_as_strings(self):
        with pytest.raises(AttributeError):
            S3CurrencyConverter('simple_string')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import with_statement, print_function, division, unicode_literals

import sys
import os.path as op
from functools import wraps
import datetime
from datetime import timedelta
from collections import defaultdict, namedtuple
from zipfile import ZipFile
from io import BytesIO
from decimal import Decimal

# We could have used "six", but like this we have no dependency
if sys.version_info[0] < 3:
    range = xrange  # noqa
    from itertools import izip as zip
    from urllib2 import urlopen

    def iteritems(d):
        return d.iteritems()

    def itervalues(d):
        return d.itervalues()


else:
    from urllib.request import urlopen

    def iteritems(d):
        return d.items()

    def itervalues(d):
        return d.values()


_DIRNAME = op.realpath(op.dirname(__file__))
CURRENCY_FILE = op.join(_DIRNAME, 'eurofxref-hist.zip')
ECB_URL = 'https://www.ecb.europa.eu/stats/eurofxref/eurofxref-hist.zip'
SINGLE_DAY_ECB_URL = 'https://www.ecb.europa.eu/stats/eurofxref/eurofxref.zip'

Bounds = namedtuple('Bounds', 'first_date last_date')

__all__ = ['CurrencyConverter',
           'S3CurrencyConverter',
           'RateNotFoundError',
           'ECB_URL',
           'SINGLE_DAY_ECB_URL', ]


def memoize(function):
    memo = {}

    @wraps(function)
    def wrapper(*args):
        if args not in memo:
            memo[args] = function(*args)
        return memo[args]
    return wrapper


@memoize
def list_dates_between(first_date, last_date):
    """Returns all dates from first to last included."""
    return [first_date + timedelta(days=n)
            for n in range(1 + (last_date - first_date).days)]


@memoize
def parse_date(s):
    """Fast %Y-%m-%d parsing."""
    try:
        return datetime.date(int(s[:4]), int(s[5:7]), int(s[8:10]))
    except ValueError:  # other accepted format used in one-day data set
        return datetime.datetime.strptime(s, '%d %B %Y').date()


def get_lines_from_zip(zip_str):
    zip_file = ZipFile(BytesIO(zip_str))
    for name in zip_file.namelist():
        for line in zip_file.read(name).decode('utf-8').splitlines():
            yield line


class RateNotFoundError(Exception):
    """Custom exception when data is missing in the rates file."""
    pass


class CurrencyConverter(object):
    """
    At init, load the historic currencies (since 1999) from the ECB.
    The rates are EUR foreign exchange reference rates:

    Date,USD,JPY,BGN,CYP,CZK,...
    2014-03-28,1.3759,140.9,1.9558,N/A,27.423,...
    2014-03-27,1.3758,...

    ``_rates`` is a dictionary with:

    - currencies as keys
    - {date: rate, ...} as values.

    ``currencies`` is a set of all available currencies.
    ``bounds`` is a dict if first and last date available per currency.
    """
    def __init__(self,
                 currency_file=CURRENCY_FILE,
                 fallback_on_wrong_date=False,
                 fallback_on_missing_rate=False,
                 fallback_on_missing_rate_method="linear_interpolation",
                 ref_currency='EUR',
                 na_values=frozenset(['', 'N/A']),
                 decimal=False,
                 verbose=False):
        """Instantiate a CurrencyConverter.

        :param str currency_file: Path to the source data. Can be a local path,
            or an URL starting with 'http://' or 'https://'. Defaults to the
            European Central Bank historical rates file included in the package.
        :param bool fallback_on_wrong_date: Set to False (default) to raise a
            RateNotFoundError when dates are requested outside the data's range.
            Set to True to extrapolate rates for dates outside the source data's
            range. The extrapolation is done by falling back to the first or
            last data point, for dates before and after the data's range,
            respectively.
        :param bool fallback_on_missing_rate: Set to True to linearly
            interpolate missing rates by their two closest valid rates. This
            only affects dates within the source data's range. Default False.
        :param bool fallback_on_missing_rate_method: Choose the fallback on missing
            rate method. Default is "linear_interpolation", also available is "last_known".
        :param str ref_currency: Three-letter currency code for the currency
            that the source data is oriented towards. This is EUR for the
            default European Central Bank data, and so the default is 'EUR'.
        :param iterable na_values: What to interpret as missing values in the
            source data.
        :param decimal: Set to True to use decimal.Decimal internally, this will
            slow the loading time but will allow exact conversions
        :param verbose: Set to True to print what is going on under the hood.
        """
        # Global options
        self.fallback_on_wrong_date = fallback_on_wrong_date
        self.fallback_on_missing_rate = fallback_on_missing_rate
        self.fallback_on_missing_rate_method = fallback_on_missing_rate_method
        self.ref_currency = ref_currency  # reference currency of rates
        self.na_values = na_values        # missing values
        self.cast = Decimal if decimal else float
        self.verbose = verbose

        # Will be filled once the file is loaded
        self._rates = None
        self.bounds = None
        self.currencies = None

        if currency_file is not None:
            self.load_file(currency_file)

    def load_file(self, currency_file):
        """To be subclassed if alternate methods of loading data.
        """
        if currency_file.startswith(('http://', 'https://')):
            content = urlopen(currency_file).read()
        else:
            with open(currency_file, 'rb') as f:
                content = f.read()

        if currency_file.endswith('.zip'):
            self.load_lines(get_lines_from_zip(content))
        else:
            self.load_lines(content.decode('utf-8').splitlines())

    def load_lines(self, lines):
        _rates = self._rates = defaultdict(dict)
        na_values = self.na_values
        cast = self.cast

        lines = iter(lines)
        header = next(lines).strip().split(',')[1:]

        for line in lines:
            line = line.strip().split(',')
            date = parse_date(line[0])
            for currency, rate in zip(header, line[1:]):
                currency = currency.strip()
                if rate not in na_values and currency:  # skip empty currency
                    _rates[currency][date] = cast(rate)

        self.currencies = set(self._rates) | set([self.ref_currency])
        self._compute_bounds()

        for currency in sorted(self._rates):
            self._set_missing_to_none(currency)
            if self.fallback_on_missing_rate:
                method = self.fallback_on_missing_rate_method
                if method == "linear_interpolation":
                    self._use_linear_interpolation(currency)
                elif method == "last_known":
                    self._use_last_known(currency)
                else:
                    raise ValueError("Unknown fallback method {0!r}".format(method))

    def _compute_bounds(self):
        self.bounds = dict((currency, Bounds(min(r), max(r)))
                           for currency, r in iteritems(self._rates))

        self.bounds[self.ref_currency] = Bounds(
            min(b.first_date for b in itervalues(self.bounds)),
            max(b.last_date for b in itervalues(self.bounds)))

    def _set_missing_to_none(self, currency):
        """Fill missing rates of a currency with the closest available ones."""
        rates = self._rates[currency]
        first_date, last_date = self.bounds[currency]

        for date in list_dates_between(first_date, last_date):
            if date not in rates:
                rates[date] = None

        if self.verbose:
            missing = len([r for r in itervalues(rates) if r is None])
            if missing:
                print('{0}: {1} missing rates from {2} to {3} ({4} days)'.format(
                    currency, missing, first_date, last_date,
                    1 + (last_date - first_date).days))

    def _use_linear_interpolation(self, currency):
        """Fill missing rates of a currency.

        This is done by linear interpolation of the two closest available rates.

        :param str currency: The currency to fill missing rates for.
        """
        rates = self._rates[currency]

        # tmp will store the closest rates forward and backward
        tmp = defaultdict(lambda: [None, None])

        for date in sorted(rates):
            rate = rates[date]
            if rate is not None:
                closest_rate = rate
                dist = 0
            else:
                dist += 1
                tmp[date][0] = closest_rate, dist

        for date in sorted(rates, reverse=True):
            rate = rates[date]
            if rate is not None:
                closest_rate = rate
                dist = 0
            else:
                dist += 1
                tmp[date][1] = closest_rate, dist

        for date in sorted(tmp):
            (r0, d0), (r1, d1) = tmp[date]
            rates[date] = (r0 * d1 + r1 * d0) / (d0 + d1)
            if self.verbose:
                print(('{0}: filling {1} missing rate using {2} ({3}d old) and '
                       '{4} ({5}d later)').format(currency, date, r0, d0, r1, d1))

    def _use_last_known(self, currency):
        """Fill missing rates of a currency.

        This is done by using the last known rate.

        :param str currency: The currency to fill missing rates for.
        """
        rates = self._rates[currency]

        for date in sorted(rates):
            rate = rates[date]
            if rate is not None:
                last_rate, last_date = rate, date
            else:
                rates[date] = last_rate
                if self.verbose:
                    print('{0}: filling {1} missing rate using {2} from {3}'.format(
                        currency, date, last_rate, last_date))

    def _get_rate(self, currency, date):
        """Get a rate for a given currency and date.

        :type date: datetime.date

        >>> from datetime import date
        >>> c = CurrencyConverter()
        >>> c._get_rate('USD', date=date(2014, 3, 28))
        1.375...
        >>> c._get_rate('BGN', date=date(2010, 11, 21))
        Traceback (most recent call last):
        RateNotFoundError: BGN has no rate for 2010-11-21
        """
        if currency == self.ref_currency:
            return self.cast('1')

        if date not in self._rates[currency]:
            first_date, last_date = self.bounds[currency]

            if not self.fallback_on_wrong_date:
                raise RateNotFoundError('{0} not in {1} bounds {2}/{3}'.format(
                    date, currency, first_date, last_date))

            if date < first_date:
                fallback_date = first_date
            elif date > last_date:
                fallback_date = last_date
            else:
                raise AssertionError('Should never happen, bug in the code!')

            if self.verbose:
                print(r'/!\ {0} not in {1} bounds {2}/{3}, falling back to {4}'.format(
                    date, currency, first_date, last_date, fallback_date))

            date = fallback_date

        rate = self._rates[currency][date]
        if rate is None:
            raise RateNotFoundError('{0} has no rate for {1}'.format(currency, date))
        return rate

    def convert(self, amount, currency, new_currency='EUR', date=None):
        """Convert amount from a currency to another one.

        :param float amount: The amount of `currency` to convert.
        :param str currency: The currency to convert from.
        :param str new_currency: The currency to convert to.
        :param datetime.date date: Use the conversion rate of this date. If this
            is not given, the most recent rate is used.

        :return: The value of `amount` in `new_currency`.
        :rtype: float

        >>> from datetime import date
        >>> c = CurrencyConverter()
        >>> c.convert(100, 'EUR', 'USD', date=date(2014, 3, 28))
        137.5...
        >>> c.convert(100, 'USD', date=date(2014, 3, 28))
        72.67...
        >>> c.convert(100, 'BGN', date=date(2010, 11, 21))
        Traceback (most recent call last):
        RateNotFoundError: BGN has no rate for 2010-11-21
        """
        for c in currency, new_currency:
            if c not in self.currencies:
                raise ValueError('{0} is not a supported currency'.format(c))

        if date is None:
            date = self.bounds[currency].last_date
        else:
            try:
                date = date.date()  # fallback if input was a datetime object
            except AttributeError:
                pass

        r0 = self._get_rate(currency, date)
        r1 = self._get_rate(new_currency, date)

        return self.cast(amount) / r0 * r1


class S3CurrencyConverter(CurrencyConverter):
    """
    Load the ECB CSV file from an S3 key instead of from a local file.
    The first argument should be an instance of boto.s3.key.Key (or any other
    object that provides a get_contents_as_string() method which returns the
    CSV file as a string).
    """
    def __init__(self, currency_file, **kwargs):
        """Make currency_file a required attribute"""
        super(S3CurrencyConverter, self).__init__(currency_file, **kwargs)

    def load_file(self, currency_file):
        lines = currency_file.get_contents_as_string().splitlines()
        self.load_lines(lines)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import argparse

from .currency_converter import CurrencyConverter, CURRENCY_FILE, parse_date


if sys.version_info[0] < 3:
    from itertools import izip_longest as zip_longest
else:
    from itertools import zip_longest


def grouper(iterable, n, fillvalue=None):
    """Group iterable by n elements.

    >>> for t in grouper('abcdefg', 3, fillvalue='x'):
    ...     print(''.join(t))
    abc
    def
    gxx
    """
    return list(zip_longest(*[iter(iterable)] * n, fillvalue=fillvalue))


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('amount', type=float)
    parser.add_argument('currency')

    parser.add_argument(
        '-t', '--to',
        help='target currency, default is %(default)s',
        default='EUR')

    parser.add_argument(
        '-d', '--date',
        help='date of rate, with format %%Y-%%m-%%d',
        default=None)

    parser.add_argument(
        '-v', '--verbose',
        help=('display available currencies, use twice (-vv) to '
              'also display details of missing rates completion'),
        action='count',
        default=0)

    parser.add_argument(
        '--decimal',
        help='use decimal.Decimal internally',
        action='store_true')

    parser.add_argument(
        '-f', '--file',
        help='change currency file used, default is %(default)s',
        default=CURRENCY_FILE)

    args = parser.parse_args()

    c = CurrencyConverter(currency_file=args.file,
                          fallback_on_wrong_date=True,
                          fallback_on_missing_rate=True,
                          decimal=args.decimal,
                          verbose=args.verbose > 1)
    currencies = sorted(c.currencies)

    if args.verbose:
        print('{0} available currencies:'.format(len(currencies)))
        for group in grouper(currencies, 10, fillvalue=''):
            print(' '.join(group))
        print('')

        currencies.sort(key=lambda u: c.bounds[u].last_date, reverse=True)
        currencies.sort(key=lambda u: c.bounds[u].first_date)
        for currency in currencies:
            first_date, last_date = c.bounds[currency]
            print('{0}: from {1} to {2} ({3} days)'.format(
                currency, first_date, last_date,
                1 + (last_date - first_date).days))
        print('')

    if args.currency not in c.currencies:
        print(r'/!\ "{0}" is not in available currencies:'.format(args.currency))
        for group in grouper(currencies, 10, fillvalue=''):
            print(' '.join(group))
        exit(1)

    if args.date is not None:
        date = parse_date(args.date)
    else:
        date = c.bounds[args.currency].last_date

    new_amount = c.convert(amount=args.amount,
                           currency=args.currency,
                           new_currency=args.to,
                           date=date)

    print('{0:,.3f} {1} = {2:,.3f} {3} on {4}'.format(
        args.amount,
        args.currency,
        new_amount,
        args.to,
        date))


if __name__ == '__main__':
    main()
